type t = int

let initial = 0

let b_flag = 1 lsl 0
let set_b context = context lor b_flag
let clr_b context = context land (lnot b_flag)
let b_set context = (context land b_flag) <> 0

let em_mask = 3
let em_start = 1
let em_smask = em_mask lsl em_start
let set_em context em = (context land (lnot em_smask)) lor (em lsl em_start)
let get_em context = (context land em_smask) lsr em_start

let i_flag = 1 lsl 3
let set_i context = context lor i_flag
let clr_i context = context land (lnot i_flag)
let i_set context = (context land i_flag) <> 0

let tt_flag = 1 lsl 4
let set_tt context = context lor tt_flag
let clr_tt context = context land (lnot tt_flag)
let tt_set context = (context land tt_flag) <> 0

let u_mask = 3
let u_start = 5
let u_smask = u_mask lsl u_start

let get_u context = (context land u_smask) lsr u_start
let inc_u context =
  if get_u context = 3 then context else context + (1 lsl u_start)
let dec_u context =
  if get_u context = 0 then context else context - (1 lsl u_start)

let sz_known_flag = 1 lsl 7
let sz_mask = 15
let sz_start = 8
let sz_smask = sz_mask lsl sz_start

let set_sz context = function
    -1 -> (context land (lnot (sz_known_flag lor sz_smask)))
  | sz -> (context land (lnot sz_smask)) lor (sz lsl sz_start) lor sz_known_flag

let get_sz context =
  if (context land sz_known_flag) = 0 then
    -1
  else
    (context land sz_smask) lsr sz_start

let cl_known_flag = 1 lsl 12
let cl_mask = 7
let cl_start = 13
let cl_smask = cl_mask lsl cl_start

let set_cl context = function
    '_' -> (context land (lnot (cl_known_flag lor cl_smask)))
  | cl ->
      (context land (lnot cl_smask))
	lor ((Color.to_int cl) lsl cl_start)
	lor cl_known_flag

let get_cl context =
  if (context land cl_known_flag) = 0 then
    Color.unknown
  else
    Color.of_int ((context land cl_smask) lsr cl_start)

let masked_by_pl = lnot (u_smask lor em_smask lor i_flag lor b_flag lor tt_flag)
let apply_pl context =
  context land masked_by_pl

let apply_tag context = function
    Tag.B -> set_b context
  | Tag.I -> set_i context
  | Tag.PL -> apply_pl context
  | Tag.TT -> set_tt context
  | Tag.U -> inc_u context
  | Tag.EM n -> set_em context n
  | Tag.SIZE s -> set_sz context s
  | Tag.COLOR c -> set_cl context c

let print fmt c =
  Format.fprintf fmt
    "@[{B<%b>@ EM<%d>@ I<%b>@ TT<%b>@ U<%d>@ SZ<%d> CL<%c>}@]"
    (b_set c) (get_em c) (i_set c) (tt_set c) (get_u c) (get_sz c) (get_cl c)

let print_std = print Format.std_formatter

let space_context context =
  let ctxt1 = clr_b (clr_i (set_em 0 context)) in
    if (get_u ctxt1 = 0) then 
      set_cl ctxt1 'w' 
    else ctxt1
;;

