type context = { b:bool; i:bool; tt:bool; u:int; em:int; size:int; color:char }
type t = context

(* new_context : context -> tag -> context option *)
let new_context context = function
  | Tag.PL -> 
      if ((context.u != 0) || (context.em>0) || (context.i) || (context.b) || (context.tt)) then Some({context with u=0;em=0;i=false;b=false;tt=false})
      else None
  | Tag.B -> 
      if context.b then None
      else Some({context with b=true})
  | Tag.I ->
      if context.i then None
      else Some({context with i=true})
  | Tag.TT -> 
      if context.tt then None
      else Some({context with tt=true})
  | Tag.U -> 
      if(context.u < 3) then Some({context with u=context.u+1})
      else None
(*
  | Tag.EM 0 -> 
      if(context.s) then None
      else 
	if(context.em) then Some({context with em=false})
	else None
  | Tag.EM 1 ->
      if(context.s) then None
      else
	if(context.em) then None
	else Some({context with em=true})
  | Tag.EM _ -> (* should be EM 2 *) 
      if(context.s) then None (* here, we know that em=false *)
      else Some({context with em=false;s=true})
*)
  | Tag.EM n -> 
      if(context.em = n) then None
      else Some({context with em=n})
  | Tag.SIZE n -> 
      if(context.size = n) then None
      else Some({context with size=n})
  | Tag.COLOR c -> 
      if(context.color = c) then None
      else Some({context with color=c})

let new_space_context context = 
  if context.u = 0 then
    {context with em=0;i=false;b=false;color='w'}
  else
    {context with em=0;i=false;b=false}

(* get_tag : stream -> (tag,bool) *)
let get_tag stream context =
  let check_next_char c = 
    if(Stream.next stream) != c then failwith "Invalid tag"
  in
  let rec get_tag_aux first_time = 
    let result =
      let c=Stream.next stream in
	match c with
	  | 'B' -> Tag.B,first_time
	  | 'E' -> 
	      check_next_char 'M';
	      if context.em = 2 then
		Tag.EM(2),first_time
	      else
		if context.em = 1 then
		  Tag.EM(0),first_time
		else
		  Tag.EM(1),first_time
	  | 'I' -> Tag.I,first_time
	  | 'P' -> check_next_char 'L';Tag.PL,first_time
	  | 'S' -> Tag.EM(2),first_time
	  | 'T' -> check_next_char 'T';Tag.TT,first_time
	  | 'U' -> Tag.U,first_time
	  | 'r' 
	  | 'g'
	  | 'b'
	  | 'c'
	  | 'm'
	  | 'y'
	  | 'k'
	  | 'w' -> Tag.COLOR(c),first_time
	  | '0'
	  | '1'
	  | '2'
	  | '3'
	  | '4'
	  | '5'
	  | '6'
	  | '7'
	  | '8'
	  | '9' -> Tag.SIZE((int_of_char c)-(int_of_char '0')),first_time
	  | '/' -> if not first_time then get_tag_aux true else failwith "Invalid tag"
	  | _ -> failwith "Invalid tag"
    in
      if not first_time then check_next_char '>';
      result
  in
    get_tag_aux false
;;

let root_context = {u=0;b=false;em=0;i=false;tt=false;size=(-1);color='_'};;


(*  (* Parse : stream -> (string,tree.t list) *) *)
(*  let parse stream = *)
(*    let follow_space=ref false  *)
(*    and last_context=ref root_context *)
(*    and start_of_tag=ref false (* !!! *) *)
(*    and spaces_to_add=ref None (* string * context *) *)
(*    and context_stack=ref [root_context] (* start * context : top is current context *) *)
(*    and child_stack=ref [] (* tree.t list : top is nested context *) *)
(*    and buffer=Buffer.create 10000 *)
(*    in *)
(*      try  *)
(*        while true do *)
(*  	let c=Stream.next stream in *)
(*  	  match c with *)
(*  	    | '<' -> *)
(*  	    | ' '  *)
(*  	    | '\013'  *)
(*  	    | '\n'  *)
(*  	    | '\t' ->  *)
(*  		(match !spaces_to_add with *)
(*  		   | Some(spaces,ctxt) *)
(*  	    | _ -> *)
(*  		(match !spaces_to_add with *)
(*  		  | Some(spaces,_) ->  *)
(*  		      Buffer.add_string buffer spaces; *)
(*  		      spaces_to_add := None; *)
(*  		  | _ -> ()); *)
(*  		Buffer.add_char buffer c  *)
(*        done; *)
(*      with Stream.Failure -> (* !!! ajouter les espaces *) Buffer.contents  *)

let private_parse stream = 
  let buffer = Buffer.create 10000 in
  let space_context = ref None in
  let rec parse_aux childs start previous context = 
(* previous is a bool option, true means char, false means whitespace *)
(* context is the current context *)
(* space_context is Some of the current space context or None *)
    try
      begin
	let c=Stream.next stream in
	  match c with
	    | '<' ->
		let (tag,closing)=get_tag stream context
		and pos=Buffer.length buffer in
		  if closing then 
		    match previous with
		      | None -> List.rev childs
		      | Some(true) -> List.rev ((Tree.Chars(start,pos))::childs)
		      | Some(false) -> List.rev ((Tree.Space(start,pos))::childs)
		  else
		    begin
		      let new_ctxt=new_context context tag in
		      let children=parse_aux [] pos None 
				     (match new_ctxt with 
					 | None -> context
					 | Some(ctxt) -> ctxt) 
		      in
		      let node=Tree.Node(tag,children) in
		      let restart=Buffer.length buffer in
			match previous with
			  | None -> parse_aux (node::childs) restart None context
			  | Some(true) -> parse_aux (node::(Tree.Chars(start,pos))::childs) restart None context
			  | Some(false) -> parse_aux (node::(Tree.Space(start,pos))::childs) restart None context
		    end
	    | ' ' 
	    | '\013' 
	    | '\n' 
	    | '\t' ->
		let old_space_context = !space_context 
		and new_space_context = new_space_context context in
		  space_context := Some(new_space_context);
		  if context.tt then 
		    begin
		      Buffer.add_char buffer c;
		      parse_aux childs start 
			(match previous with
			   | None -> (Some(false))
			   | Some(_) -> previous) context
		    end
		  else
		    begin
		      match old_space_context with
			| None ->
			    begin
			      Buffer.add_char buffer c;
			      parse_aux childs start 
				(match previous with
				   | None -> (Some(false))
				   | Some(_) -> previous) context
			    end
			| Some(spc_ctxt) ->
			    begin
			      if(spc_ctxt = new_space_context) then
				begin
				  parse_aux childs start previous context
				end
			      else
				begin
				  Buffer.add_char buffer c;
				  parse_aux childs start 
				    (match previous with
				       | None -> (Some(false))
				       | Some(_) -> previous) context
				end
			    end
		    end
	    | _ ->
		space_context := None;
		Buffer.add_char buffer c;
		parse_aux childs start (Some(true)) context 
      end
    with
	Stream.Failure -> 
	  let stop=Buffer.length buffer in
	    match previous with
	      | None -> List.rev childs
	      | Some(true) -> List.rev ((Tree.Chars(start,stop))::childs)
	      | Some(false) -> List.rev ((Tree.Space(start,stop))::childs)
  in
  let result=parse_aux [] 0 None root_context in
    (Buffer.contents buffer,result)
;;

let parse_with_stack stream =
  let buffer = Buffer.create 10000 in
  let stack=Stack.create () in
  let previous = ref None 
  and space_context = ref None
  and context = ref root_context 
  and start = ref 0 
  and children = ref [] in
    try
      while true do
	let c=Stream.next stream in
	  match c with
	    | '<' ->
		let (tag,closing)=get_tag stream (!context) 
		and pos=Buffer.length buffer in
		  if closing then
		    begin
		      let (tag,ctxt,brothers,significant)=Stack.pop stack in
			if significant then
			  (* here, the tag was significant *)
			  begin
			    (match (!previous) with
			       | None -> ()
			       | Some(true) -> children := (Tree.Chars(!start,pos)) :: (!children)
			       | Some(false) -> children := (Tree.Space(!start,pos)) :: (!children));
			    let node=Tree.Node(tag,List.rev (!children)) in
			      children := node :: brothers;
			      previous := None;
			      start := pos;
			      context := ctxt
			  end
			else
			  (* here, the tag can be deleted *)
			  begin
			    (*
			      (match (!previous) with
			      | None -> ()
			      | Some(true) -> children := (Tree.Chars(!start,pos)) :: (!children)
			      | Some(false) -> children := (Tree.Space(!start,pos)) :: (!children));
			      let node=Tree.Node(tag,List.rev (!children)) in
			      children := node :: brothers;
			      previous := None;
			      start := pos;
			      context := ctxt
			    *)
			    context := ctxt
			  end
		    end
		  else
		    begin
		      let new_ctxt=new_context (!context) tag in
			match new_ctxt with
			  | None ->
			      (* here, the tag can be deleted *)
			      begin
				(*
				  (match (!previous) with
				  | None -> ()
				  | Some(true) -> children := (Tree.Chars(!start,pos)) :: (!children)
				  | Some(false) -> children := (Tree.Space(!start,pos)) :: (!children));
				  Stack.push (tag,(!context),(!children),false) stack;
				  children := [];
				  previous := None;
				  start := pos;
				*)
				Stack.push (tag,(!context),(!children),false) stack;
			      end
			  | Some(ctxt) ->
			      (* here, the tag is significant *)
			      begin
				(match (!previous) with
				   | None -> ()
				   | Some(true) -> children := (Tree.Chars(!start,pos)) :: (!children)
				   | Some(false) -> children := (Tree.Space(!start,pos)) :: (!children));
				Stack.push (tag,(!context),(!children),true) stack;
				context := ctxt;
				children := [];
				previous := None;
				start := pos;
			      end
		    end
	    | ' ' 
	    | '\013' 
	    | '\n' 
	    | '\t' ->
		begin
		  let old_space_context = !space_context 
		  and new_space_context = new_space_context (!context) in
		    space_context := Some(new_space_context);
		    if (!context).tt then
		      begin
			Buffer.add_char buffer c;
			match (!previous) with
			  | None -> previous := Some(false)
			  | _ -> ()
		      end
		    else
		      begin
			match old_space_context with
			  | None ->
			      begin
				Buffer.add_char buffer c;
				match (!previous) with
				  | None -> previous := Some(false)
				  | _ -> ()
			      end
			  | Some(spc_ctxt) ->
			      if not (spc_ctxt = new_space_context) then
				begin
				  Buffer.add_char buffer c;
				  match (!previous) with
				    | None -> previous := Some(false)
				    | _ -> ()
				end
		      end
		end
	    | _ -> 
		begin
		  Buffer.add_char buffer c;
		  previous := Some(true);
		  space_context := None;
		end
      done;
      (Buffer.contents buffer,List.rev (!children));
    with
      | Stream.Failure ->
	  begin
	    let pos=Buffer.length buffer in
	      (match (!previous) with
		 | None -> ()
		 | Some(true) -> children := (Tree.Chars(!start,pos)) :: (!children)
		 | Some(false) -> children := (Tree.Space(!start,pos)) :: (!children));
	      (Buffer.contents buffer,List.rev (!children))
	  end
;;

let parse = parse_with_stack;;




